import itertools
from contextlib import ExitStack
from typing import Iterator, Mapping, MutableSet, Optional, Tuple, Union, cast

from afancontrol.config import AnyFanName, FanName, ReadonlyFanName
from afancontrol.logger import logger
from afancontrol.pwmfannorm import PWMFanNorm, PWMValueNorm, ReadonlyPWMFanNorm
from afancontrol.report import Report


class Fans:
    def __init__(
        self,
        *,
        fans: Mapping[FanName, PWMFanNorm],
        readonly_fans: Mapping[ReadonlyFanName, ReadonlyPWMFanNorm],
        report: Report
    ) -> None:
        self.fans = fans
        self.readonly_fans = readonly_fans
        self.report = report
        self._stack = None  # type: Optional[ExitStack]

        # Set of fans marked as failing (which speed is 0)
        self._failed_fans = set()  # type: MutableSet[AnyFanName]

        # Set of fans that will be skipped on speed check
        self._stopped_fans = set()  # type: MutableSet[AnyFanName]

    def is_fan_failing(self, fan_name: AnyFanName) -> bool:
        return fan_name in self._failed_fans

    def is_fan_stopped(self, fan_name: AnyFanName) -> bool:
        return fan_name in self._stopped_fans

    def __enter__(self):  # reusable
        self._stack = ExitStack()
        logger.info("Enabling PWM on fans...")
        try:
            for pwmfan in cast(
                Iterator[Union[PWMFanNorm, ReadonlyPWMFanNorm]],
                itertools.chain(self.fans.values(), self.readonly_fans.values()),
            ):
                self._stack.enter_context(pwmfan)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self._stack is not None
        logger.info("Disabling PWM on fans...")
        self._stack.close()
        logger.info("Done. Fans should be returned to full speed")
        return None

    def check_speeds(self) -> None:
        for name, fan in cast(
            Iterator[Tuple[AnyFanName, Union[PWMFanNorm, ReadonlyPWMFanNorm]]],
            itertools.chain(self.fans.items(), self.readonly_fans.items()),
        ):
            if name in self._stopped_fans:
                continue
            try:
                if fan.get_speed() <= 0:
                    raise RuntimeError("Fan speed is 0")
            except Exception as e:
                self._ensure_fan_is_failing(name, e)
            else:
                self._ensure_fan_is_not_failing(name)

    def set_all_to_full_speed(self) -> None:
        for name, fan in self.fans.items():
            if name in self._failed_fans:
                continue
            try:
                fan.set_full_speed()
            except Exception as e:
                logger.warning("Unable to set the fan '%s' to full speed:\n%s", name, e)

    def set_fan_speeds(self, speeds: Mapping[FanName, PWMValueNorm]) -> None:
        assert speeds.keys() == self.fans.keys()
        self._stopped_fans.clear()
        for name, pwm_norm in speeds.items():
            fan = self.fans[name]
            assert 0.0 <= pwm_norm <= 1.0

            if name in self._failed_fans:
                continue

            try:
                pwm = fan.set(pwm_norm)
            except Exception as e:
                logger.warning(
                    "Unable to set the fan '%s' to speed %s:\n%s", name, pwm_norm, e
                )
            else:
                logger.debug(
                    "Fan status [%s]: speed: %.3f, pwm: %s", name, pwm_norm, pwm
                )
                if fan.is_pwm_stopped(pwm):
                    self._stopped_fans.add(name)
        for readonly_name, readonly_fan in self.readonly_fans.items():
            readonly_pwm_norm = readonly_fan.get()
            readonly_pwm = readonly_fan.get_raw()
            logger.debug(
                "Readonly Fan status [%s]: speed: %.3f, pwm: %s",
                readonly_name,
                readonly_pwm_norm,
                readonly_pwm,
            )
            if readonly_fan.is_pwm_stopped(readonly_pwm):
                self._stopped_fans.add(readonly_name)

    def _ensure_fan_is_failing(
        self, name: AnyFanName, get_speed_exc: Exception
    ) -> None:
        if name in self._failed_fans:
            return
        self._failed_fans.add(name)
        try:
            fan = self.fans[cast(FanName, name)]
        except KeyError:
            self.readonly_fans[cast(ReadonlyFanName, name)]  # assert
            full_speed_result = "The fan is in the readonly mode"
        else:
            try:
                # Perhaps it had jammed, so setting it to full speed might
                # recover it?
                fan.set_full_speed()
            except Exception as e:
                full_speed_result = "Setting fan speed to full has failed:\n%s" % e
            else:
                full_speed_result = "Fan has been set to full speed"

        self.report.report(
            "fan stopped: %s" % name,
            "Looks like the fan '%s' is failing:\n%s\n\n%s"
            % (name, get_speed_exc, full_speed_result),
        )

    def _ensure_fan_is_not_failing(self, name: AnyFanName) -> None:
        if name not in self._failed_fans:
            return
        self.report.report(
            "fan started: %s" % name,
            "Fan '%s' which had previously been reported as failing has just started."
            % name,
        )
        self._failed_fans.remove(name)



import csv
import io

from afancontrol.exec import exec_shell_command
from afancontrol.pwmfan.base import BaseFanSpeed, FanValue

# TODO maybe switch to `python3-pyghmi`? although it looks like the current version
# in Stretch is broken in py3: https://opendev.org/x/pyghmi/commit/2e12f5ce15e11e46a1c11ee3b00b94cb8bd7feb9  # noqa


class FreeIPMIFanSpeed(BaseFanSpeed):
    __slots__ = ("_name", "_ipmi_sensors_bin", "_ipmi_sensors_extra_args")

    def __init__(
        self, name: str, *, ipmi_sensors_bin="ipmi-sensors", ipmi_sensors_extra_args=""
    ) -> None:
        self._name = name
        self._ipmi_sensors_bin = ipmi_sensors_bin
        self._ipmi_sensors_extra_args = ipmi_sensors_extra_args

    def get_speed(self) -> FanValue:
        out = self._call_ipmi_sensors()
        reader = csv.DictReader(io.StringIO(out))
        for row in reader:
            if row["Name"] == self._name:
                assert row["Units"] == "RPM"
                # assert row["Event"] == "'OK'"
                return FanValue(int(float(row["Reading"])))
        raise RuntimeError(
            "ipmi-sensors output doesn't contain %r fan:\n%s" % (self._name, out)
        )

    def _call_ipmi_sensors(self) -> str:
        shell_command = "%s %s --sensor-types Fan --comma-separated-output" % (
            self._ipmi_sensors_bin,
            self._ipmi_sensors_extra_args,
        )
        return exec_shell_command(shell_command, timeout=2)



import pytest

from afancontrol.temp import FileTemp, TempCelsius, TempStatus


@pytest.fixture
def file_temp_path(temp_path):
    # /sys/class/hwmon/hwmon0/temp1_input
    temp_input_path = temp_path / "temp1_input"
    temp_input_path.write_text("34000\n")

    temp_max_path = temp_path / "temp1_max"
    temp_max_path.write_text("127000\n")

    temp_min_path = temp_path / "temp1_min"
    # My mobo actually returns this as min:
    temp_min_path.write_text("127000\n")

    return temp_input_path


def test_file_temp_min_max_numbers(file_temp_path):
    temp = FileTemp(
        temp_path=str(file_temp_path),
        min=TempCelsius(40.0),
        max=TempCelsius(50.0),
        panic=TempCelsius(60.0),
        threshold=None,
    )
    assert temp.get() == TempStatus(
        temp=TempCelsius(34.0),
        min=TempCelsius(40.0),
        max=TempCelsius(50.0),
        panic=TempCelsius(60.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )
    print(repr(temp))


def test_file_temp_glob(file_temp_path):
    temp = FileTemp(
        temp_path=str(file_temp_path).replace("/temp1", "/temp?"),
        min=TempCelsius(40.0),
        max=None,
        panic=None,
        threshold=None,
    )
    assert temp.get() == TempStatus(
        temp=TempCelsius(34.0),
        min=TempCelsius(40.0),
        max=TempCelsius(127.0),
        panic=None,
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )
    print(repr(temp))


def test_file_temp_min_max_files(temp_path, file_temp_path):
    with pytest.raises(RuntimeError):
        # min == max is an error
        FileTemp(
            temp_path=str(file_temp_path),
            min=None,
            max=None,
            panic=TempCelsius(60.0),
            threshold=None,
        ).get()

    temp = FileTemp(
        temp_path=str(file_temp_path),
        min=TempCelsius(50.0),
        max=None,
        panic=TempCelsius(60.0),
        threshold=None,
    )
    assert temp.get() == TempStatus(
        temp=TempCelsius(34.0),
        min=TempCelsius(50.0),
        max=TempCelsius(127.0),
        panic=TempCelsius(60.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )



from typing import Optional, Tuple

from afancontrol.exec import exec_shell_command
from afancontrol.temp.base import Temp, TempCelsius


def _is_float(s: str) -> bool:
    if not s:
        return False
    try:
        float(s)
    except (ValueError, TypeError):
        return False
    else:
        return True


class HDDTemp(Temp):
    def __init__(
        self,
        disk_path: str,
        *,
        min: TempCelsius,
        max: TempCelsius,
        panic: Optional[TempCelsius],
        threshold: Optional[TempCelsius],
        hddtemp_bin: str = "hddtemp"
    ) -> None:
        super().__init__(panic=panic, threshold=threshold)
        self._disk_path = disk_path
        self._min = min
        self._max = max
        self._hddtemp_bin = hddtemp_bin

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._disk_path == other._disk_path
                and self._min == other._min
                and self._max == other._max
                and self._panic == other._panic
                and self._threshold == other._threshold
                and self._hddtemp_bin == other._hddtemp_bin
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, min=%r, max=%r, panic=%r, threshold=%r, hddtemp_bin=%r)" % (
            type(self).__name__,
            self._disk_path,
            self._min,
            self._max,
            self._panic,
            self._threshold,
            self._hddtemp_bin,
        )

    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        temps = [
            float(line.strip())
            for line in self._call_hddtemp().split("\n")
            if _is_float(line.strip())
        ]
        if not temps:
            raise RuntimeError(
                "hddtemp returned empty list of valid temperature values"
            )
        temp = TempCelsius(max(temps))
        return temp, self._get_min(), self._get_max()

    def _get_min(self) -> TempCelsius:
        return TempCelsius(self._min)

    def _get_max(self) -> TempCelsius:
        return TempCelsius(self._max)

    def _call_hddtemp(self) -> str:
        # `disk_path` might be a glob, so it has to be executed with a shell.
        shell_command = "%s -n -u C -- %s" % (self._hddtemp_bin, self._disk_path)
        return exec_shell_command(shell_command, timeout=10)



from afancontrol.temp import CommandTemp, TempCelsius, TempStatus


def test_command_temp_with_minmax():
    t = CommandTemp(
        shell_command=r"printf '%s\n' 35 30 40",
        min=TempCelsius(31.0),
        max=TempCelsius(39.0),
        panic=TempCelsius(50.0),
        threshold=None,
    )
    assert t.get() == TempStatus(
        temp=TempCelsius(35.0),
        min=TempCelsius(31.0),
        max=TempCelsius(39.0),
        panic=TempCelsius(50.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )
    print(repr(t))


def test_command_temp_without_minmax():
    t = CommandTemp(
        shell_command=r"printf '%s\n' 35 30 40",
        min=None,
        max=None,
        panic=TempCelsius(50.0),
        threshold=None,
    )
    assert t.get() == TempStatus(
        temp=TempCelsius(35.0),
        min=TempCelsius(30.0),
        max=TempCelsius(40.0),
        panic=TempCelsius(50.0),
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )



import abc
import collections
from typing import Deque, Optional, TypeVar

from afancontrol.temp import TempStatus

T = TypeVar("T")


class TempFilter(abc.ABC):
    @abc.abstractmethod
    def copy(self: T) -> T:
        pass

    @abc.abstractmethod
    def apply(self, status: Optional[TempStatus]) -> Optional[TempStatus]:
        pass

    def __enter__(self):  # reusable
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass


class NullFilter(TempFilter):
    def copy(self: T) -> T:
        return type(self)()

    def apply(self, status: Optional[TempStatus]) -> Optional[TempStatus]:
        return status

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return True

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s()" % (type(self).__name__,)


def _temp_status_sorting_key(status: Optional[TempStatus]) -> float:
    if status is None:
        return float("+inf")
    return status.temp


class MovingQuantileFilter(TempFilter):
    def __init__(self, quantile: float, *, window_size: int) -> None:
        self.quantile = quantile
        self.window_size = window_size
        self.history = None  # type: Optional[Deque[Optional[TempStatus]]]

    def copy(self: T) -> T:
        return type(self)(  # type: ignore
            quantile=self.quantile, window_size=self.window_size  # type: ignore
        )

    def apply(self, status: Optional[TempStatus]) -> Optional[TempStatus]:
        assert self.history is not None
        self.history.append(status)

        observations = sorted(self.history, key=_temp_status_sorting_key)
        target_idx = int(len(observations) * self.quantile)
        return observations[target_idx]

    def __enter__(self):  # reusable
        assert self.history is None
        self.history = collections.deque(maxlen=self.window_size)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self.history is not None
        self.history = None

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.quantile == other.quantile
                and self.window_size == other.window_size
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(quantile=%r, window_size=%r)" % (
            type(self).__name__,
            self.quantile,
            self.window_size,
        )


class MovingMedianFilter(MovingQuantileFilter):
    def __init__(self, window_size: int) -> None:
        super().__init__(quantile=0.5, window_size=window_size)

    def copy(self: T) -> T:
        return type(self)(window_size=self.window_size)  # type: ignore



import pytest

from afancontrol.filters import MovingMedianFilter, MovingQuantileFilter, NullFilter
from afancontrol.temp import TempCelsius, TempStatus


def make_temp_status(temp):
    return TempStatus(
        min=TempCelsius(30),
        max=TempCelsius(50),
        temp=TempCelsius(temp),
        panic=None,
        threshold=None,
        is_panic=False,
        is_threshold=False,
    )


@pytest.mark.parametrize(
    "filter",
    [
        NullFilter(),
        MovingMedianFilter(window_size=3),
        MovingQuantileFilter(0.5, window_size=3),
    ],
)
def test_none(filter):
    with filter:
        assert filter.apply(None) is None


@pytest.mark.parametrize(
    "filter",
    [
        NullFilter(),
        MovingMedianFilter(window_size=3),
        MovingQuantileFilter(0.5, window_size=3),
    ],
)
def test_single_point(filter):
    with filter:
        assert filter.apply(make_temp_status(42.0)) == make_temp_status(42.0)


def test_moving_quantile():
    f = MovingQuantileFilter(0.8, window_size=10)
    with f:
        assert f.apply(make_temp_status(42.0)) == make_temp_status(42.0)
        assert f.apply(make_temp_status(45.0)) == make_temp_status(45.0)
        assert f.apply(make_temp_status(47.0)) == make_temp_status(47.0)
        assert f.apply(make_temp_status(123.0)) == make_temp_status(123.0)
        assert f.apply(make_temp_status(46.0)) == make_temp_status(123.0)
        assert f.apply(make_temp_status(49.0)) == make_temp_status(49.0)
        assert f.apply(make_temp_status(51.0)) == make_temp_status(51.0)
        assert f.apply(None) == make_temp_status(123.0)
        assert f.apply(None) is None
        assert f.apply(make_temp_status(51.0)) is None
        assert f.apply(make_temp_status(53.0)) is None


def test_moving_median():
    f = MovingMedianFilter(window_size=3)
    with f:
        assert f.apply(make_temp_status(42.0)) == make_temp_status(42.0)
        assert f.apply(make_temp_status(45.0)) == make_temp_status(45.0)
        assert f.apply(make_temp_status(47.0)) == make_temp_status(45.0)
        assert f.apply(make_temp_status(123.0)) == make_temp_status(47.0)
        assert f.apply(make_temp_status(46.0)) == make_temp_status(47.0)
        assert f.apply(make_temp_status(49.0)) == make_temp_status(49.0)
        assert f.apply(make_temp_status(51.0)) == make_temp_status(49.0)
        assert f.apply(None) == make_temp_status(51.0)
        assert f.apply(None) is None
        assert f.apply(make_temp_status(51.0)) is None
        assert f.apply(make_temp_status(53.0)) == make_temp_status(53.0)



from pathlib import Path
from unittest.mock import Mock

import pytest

from afancontrol.arduino import (
    ArduinoConnection,
    ArduinoName,
    ArduinoPin,
    pyserial_available,
)
from afancontrol.config import (
    Actions,
    AlertCommands,
    DaemonCLIConfig,
    DaemonConfig,
    FanName,
    FanSpeedModifier,
    FansTempsRelation,
    FilteredTemp,
    MappingName,
    ParsedConfig,
    ReadonlyFanName,
    TempName,
    TriggerConfig,
    parse_config,
)
from afancontrol.filters import MovingMedianFilter, NullFilter
from afancontrol.pwmfan import (
    ArduinoFanPWMRead,
    ArduinoFanPWMWrite,
    ArduinoFanSpeed,
    FanInputDevice,
    LinuxFanPWMRead,
    LinuxFanPWMWrite,
    LinuxFanSpeed,
    PWMDevice,
    PWMValue,
)
from afancontrol.pwmfannorm import PWMFanNorm, ReadonlyPWMFanNorm
from afancontrol.temp import FileTemp, HDDTemp, TempCelsius


@pytest.fixture
def pkg_conf():
    return Path(__file__).parents[1] / "pkg" / "afancontrol.conf"


@pytest.fixture
def example_conf():
    return Path(__file__).parents[0] / "data" / "afancontrol-example.conf"


def path_from_str(contents: str) -> Path:
    p = Mock(spec=Path)
    p.read_text.return_value = contents
    return p


@pytest.mark.skipif(not pyserial_available, reason="pyserial is not installed")
def test_pkg_conf(pkg_conf: Path):
    daemon_cli_config = DaemonCLIConfig(
        pidfile=None, logfile=None, exporter_listen_host=None
    )

    parsed = parse_config(pkg_conf, daemon_cli_config)
    assert parsed == ParsedConfig(
        arduino_connections={},
        daemon=DaemonConfig(
            pidfile="/run/afancontrol.pid",
            logfile="/var/log/afancontrol.log",
            interval=5,
            exporter_listen_host=None,
        ),
        report_cmd=(
            'printf "Subject: %s\nTo: %s\n\n%b" '
            '"afancontrol daemon report: %REASON%" root "%MESSAGE%" | sendmail -t'
        ),
        triggers=TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                )
            },
        ),
        fans={
            FanName("hdd"): PWMFanNorm(
                fan_speed=LinuxFanSpeed(
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan2_input")
                ),
                pwm_read=LinuxFanPWMRead(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2")
                ),
                pwm_write=LinuxFanPWMWrite(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2")
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=False,
            )
        },
        readonly_fans={
            ReadonlyFanName("cpu"): ReadonlyPWMFanNorm(
                fan_speed=LinuxFanSpeed(
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan1_input")
                ),
            ),
        },
        temps={
            TempName("mobo"): FilteredTemp(
                temp=FileTemp(
                    "/sys/class/hwmon/hwmon0/device/temp1_input",
                    min=TempCelsius(30.0),
                    max=TempCelsius(40.0),
                    panic=None,
                    threshold=None,
                ),
                filter=MovingMedianFilter(window_size=3),
            )
        },
        mappings={
            MappingName("1"): FansTempsRelation(
                temps=[TempName("mobo")],
                fans=[FanSpeedModifier(fan=FanName("hdd"), modifier=0.6)],
            )
        },
    )


@pytest.mark.skipif(not pyserial_available, reason="pyserial is not installed")
def test_example_conf(example_conf: Path):
    daemon_cli_config = DaemonCLIConfig(
        pidfile=None, logfile=None, exporter_listen_host=None
    )

    parsed = parse_config(example_conf, daemon_cli_config)
    assert parsed == ParsedConfig(
        arduino_connections={
            ArduinoName("mymicro"): ArduinoConnection(
                ArduinoName("mymicro"), "/dev/ttyACM0", baudrate=115200, status_ttl=5
            )
        },
        daemon=DaemonConfig(
            pidfile="/run/afancontrol.pid",
            logfile="/var/log/afancontrol.log",
            exporter_listen_host="127.0.0.1:8083",
            interval=5,
        ),
        report_cmd=(
            'printf "Subject: %s\nTo: %s\n\n%b" '
            '"afancontrol daemon report: %REASON%" root "%MESSAGE%" | sendmail -t'
        ),
        triggers=TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("hdds"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                ),
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                ),
            },
        ),
        fans={
            FanName("cpu"): PWMFanNorm(
                fan_speed=LinuxFanSpeed(
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan1_input")
                ),
                pwm_read=LinuxFanPWMRead(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm1")
                ),
                pwm_write=LinuxFanPWMWrite(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm1")
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=True,
            ),
            FanName("hdd"): PWMFanNorm(
                fan_speed=LinuxFanSpeed(
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan2_input")
                ),
                pwm_read=LinuxFanPWMRead(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2")
                ),
                pwm_write=LinuxFanPWMWrite(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2")
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=False,
            ),
            FanName("my_arduino_fan"): PWMFanNorm(
                fan_speed=ArduinoFanSpeed(
                    ArduinoConnection(
                        ArduinoName("mymicro"),
                        "/dev/ttyACM0",  # linux
                        # "/dev/cu.usbmodem14201",  # macos
                        baudrate=115200,
                        status_ttl=5,
                    ),
                    tacho_pin=ArduinoPin(3),
                ),
                pwm_read=ArduinoFanPWMRead(
                    ArduinoConnection(
                        ArduinoName("mymicro"),
                        "/dev/ttyACM0",  # linux
                        # "/dev/cu.usbmodem14201",  # macos
                        baudrate=115200,
                        status_ttl=5,
                    ),
                    pwm_pin=ArduinoPin(9),
                ),
                pwm_write=ArduinoFanPWMWrite(
                    ArduinoConnection(
                        ArduinoName("mymicro"),
                        "/dev/ttyACM0",  # linux
                        # "/dev/cu.usbmodem14201",  # macos
                        baudrate=115200,
                        status_ttl=5,
                    ),
                    pwm_pin=ArduinoPin(9),
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=True,
            ),
        },
        readonly_fans={},
        temps={
            TempName("hdds"): FilteredTemp(
                temp=HDDTemp(
                    "/dev/sd?",
                    min=TempCelsius(35.0),
                    max=TempCelsius(48.0),
                    panic=TempCelsius(55.0),
                    threshold=None,
                    hddtemp_bin="hddtemp",
                ),
                filter=NullFilter(),
            ),
            TempName("mobo"): FilteredTemp(
                temp=FileTemp(
                    "/sys/class/hwmon/hwmon0/device/temp1_input",
                    min=TempCelsius(30.0),
                    max=TempCelsius(40.0),
                    panic=None,
                    threshold=None,
                ),
                filter=NullFilter(),
            ),
        },
        mappings={
            MappingName("1"): FansTempsRelation(
                temps=[TempName("mobo"), TempName("hdds")],
                fans=[
                    FanSpeedModifier(fan=FanName("cpu"), modifier=1.0),
                    FanSpeedModifier(fan=FanName("hdd"), modifier=0.6),
                    FanSpeedModifier(fan=FanName("my_arduino_fan"), modifier=0.222),
                ],
            ),
            MappingName("2"): FansTempsRelation(
                temps=[TempName("hdds")],
                fans=[FanSpeedModifier(fan=FanName("hdd"), modifier=1.0)],
            ),
        },
    )


def test_minimal_config() -> None:
    daemon_cli_config = DaemonCLIConfig(
        pidfile=None, logfile=None, exporter_listen_host=None
    )

    config = """
[daemon]

[actions]

[temp:mobo]
type = file
path = /sys/class/hwmon/hwmon0/device/temp1_input

[fan: case]
pwm = /sys/class/hwmon/hwmon0/device/pwm2
fan_input = /sys/class/hwmon/hwmon0/device/fan2_input

[mapping:1]
fans = case*0.6,
temps = mobo
"""
    parsed = parse_config(path_from_str(config), daemon_cli_config)
    assert parsed == ParsedConfig(
        arduino_connections={},
        daemon=DaemonConfig(
            pidfile="/run/afancontrol.pid",
            logfile=None,
            exporter_listen_host=None,
            interval=5,
        ),
        report_cmd=(
            'printf "Subject: %s\nTo: %s\n\n%b" '
            '"afancontrol daemon report: %REASON%" root "%MESSAGE%" | sendmail -t'
        ),
        triggers=TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                )
            },
        ),
        fans={
            FanName("case"): PWMFanNorm(
                fan_speed=LinuxFanSpeed(
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan2_input")
                ),
                pwm_read=LinuxFanPWMRead(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2")
                ),
                pwm_write=LinuxFanPWMWrite(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm2")
                ),
                pwm_line_start=PWMValue(100),
                pwm_line_end=PWMValue(240),
                never_stop=True,
            )
        },
        readonly_fans={},
        temps={
            TempName("mobo"): FilteredTemp(
                temp=FileTemp(
                    "/sys/class/hwmon/hwmon0/device/temp1_input",
                    min=None,
                    max=None,
                    panic=None,
                    threshold=None,
                ),
                filter=NullFilter(),
            )
        },
        mappings={
            MappingName("1"): FansTempsRelation(
                temps=[TempName("mobo")],
                fans=[FanSpeedModifier(fan=FanName("case"), modifier=0.6)],
            )
        },
    )


def test_readonly_config() -> None:
    daemon_cli_config = DaemonCLIConfig(
        pidfile=None, logfile=None, exporter_listen_host=None
    )

    config = """
[daemon]

[actions]

[temp:mobo]
type = file
path = /sys/class/hwmon/hwmon0/device/temp1_input

[readonly_fan: cpu]
pwm = /sys/class/hwmon/hwmon0/device/pwm1
fan_input = /sys/class/hwmon/hwmon0/device/fan1_input
"""
    parsed = parse_config(path_from_str(config), daemon_cli_config)
    assert parsed == ParsedConfig(
        arduino_connections={},
        daemon=DaemonConfig(
            pidfile="/run/afancontrol.pid",
            logfile=None,
            exporter_listen_host=None,
            interval=5,
        ),
        report_cmd=(
            'printf "Subject: %s\nTo: %s\n\n%b" '
            '"afancontrol daemon report: %REASON%" root "%MESSAGE%" | sendmail -t'
        ),
        triggers=TriggerConfig(
            global_commands=Actions(
                panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
            ),
            temp_commands={
                TempName("mobo"): Actions(
                    panic=AlertCommands(enter_cmd=None, leave_cmd=None),
                    threshold=AlertCommands(enter_cmd=None, leave_cmd=None),
                )
            },
        ),
        fans={},
        readonly_fans={
            ReadonlyFanName("cpu"): ReadonlyPWMFanNorm(
                fan_speed=LinuxFanSpeed(
                    FanInputDevice("/sys/class/hwmon/hwmon0/device/fan1_input")
                ),
                pwm_read=LinuxFanPWMRead(
                    PWMDevice("/sys/class/hwmon/hwmon0/device/pwm1")
                ),
            )
        },
        temps={
            TempName("mobo"): FilteredTemp(
                temp=FileTemp(
                    "/sys/class/hwmon/hwmon0/device/temp1_input",
                    min=None,
                    max=None,
                    panic=None,
                    threshold=None,
                ),
                filter=NullFilter(),
            )
        },
        mappings={},
    )



from collections import OrderedDict
from unittest.mock import MagicMock

import pytest

from afancontrol.config import FanName
from afancontrol.fans import Fans
from afancontrol.pwmfan import BaseFanPWMRead
from afancontrol.pwmfannorm import PWMFanNorm, PWMValueNorm
from afancontrol.report import Report


@pytest.fixture
def report():
    return MagicMock(spec=Report)


@pytest.mark.parametrize("is_fan_failing", [False, True])
def test_smoke(report, is_fan_failing):
    fan = MagicMock(spec=PWMFanNorm)
    fans = Fans(fans={FanName("test"): fan}, readonly_fans={}, report=report)

    fan.set = lambda pwm_norm: int(255 * pwm_norm)
    fan.get_speed.return_value = 0 if is_fan_failing else 942
    fan.is_pwm_stopped = BaseFanPWMRead.is_pwm_stopped

    with fans:
        assert 1 == fan.__enter__.call_count
        fans.check_speeds()
        fans.set_all_to_full_speed()
        fans.set_fan_speeds({FanName("test"): PWMValueNorm(0.42)})
        assert fan.get_speed.call_count == 1
        if is_fan_failing:
            assert fans._failed_fans == {"test"}
            assert fans._stopped_fans == set()
        else:
            assert fans._failed_fans == set()
            assert fans._stopped_fans == set()

    assert 1 == fan.__exit__.call_count


def test_set_fan_speeds(report):
    mocked_fans = OrderedDict(
        [
            (FanName("test1"), MagicMock(spec=PWMFanNorm)),
            (FanName("test2"), MagicMock(spec=PWMFanNorm)),
            (FanName("test3"), MagicMock(spec=PWMFanNorm)),
            (FanName("test4"), MagicMock(spec=PWMFanNorm)),
        ]
    )

    for fan in mocked_fans.values():
        fan.set.return_value = 240
        fan.get_speed.return_value = 942
        fan.is_pwm_stopped = BaseFanPWMRead.is_pwm_stopped

    fans = Fans(fans=mocked_fans, readonly_fans={}, report=report)
    with fans:
        fans._ensure_fan_is_failing(FanName("test2"), Exception("test"))
        fans.set_fan_speeds(
            {
                FanName("test1"): PWMValueNorm(0.42),
                FanName("test2"): PWMValueNorm(0.42),
                FanName("test3"): PWMValueNorm(0.42),
                FanName("test4"): PWMValueNorm(0.42),
            }
        )
        assert [1, 0, 1, 1] == [f.set.call_count for f in mocked_fans.values()]



from typing import Optional
from unittest.mock import patch

import pytest

from afancontrol.temp import Temp, TempCelsius, TempStatus


class DummyTemp(Temp):
    def _get_temp(self):
        pass


@pytest.mark.parametrize(
    "temp, threshold, panic, is_threshold, is_panic",
    [
        (34.0, None, 60.0, False, False),
        (42.0, None, 60.0, False, False),
        (57.0, 55.0, 60.0, True, False),
        (61.0, 55.0, 61.0, True, True),
        (61.0, None, 61.0, False, True),
    ],
)
def test_temp(
    temp: TempCelsius,
    threshold: Optional[TempCelsius],
    panic: TempCelsius,
    is_threshold,
    is_panic,
):
    min = TempCelsius(40.0)
    max = TempCelsius(50.0)

    with patch.object(DummyTemp, "_get_temp") as mock_get_temp:
        t = DummyTemp(panic=panic, threshold=threshold)
        mock_get_temp.return_value = [temp, min, max]

        assert t.get() == TempStatus(
            temp=temp,
            min=min,
            max=max,
            panic=panic,
            threshold=threshold,
            is_panic=is_panic,
            is_threshold=is_threshold,
        )



from afancontrol.pwmfan.arduino import (
    ArduinoFanPWMRead,
    ArduinoFanPWMWrite,
    ArduinoFanSpeed,
)
from afancontrol.pwmfan.base import (
    BaseFanPWMRead,
    BaseFanPWMWrite,
    BaseFanSpeed,
    FanValue,
    PWMValue,
)
from afancontrol.pwmfan.ipmi import FreeIPMIFanSpeed
from afancontrol.pwmfan.linux import (
    FanInputDevice,
    LinuxFanPWMRead,
    LinuxFanPWMWrite,
    LinuxFanSpeed,
    PWMDevice,
)

__all__ = (
    "ArduinoFanPWMRead",
    "ArduinoFanPWMWrite",
    "ArduinoFanSpeed",
    "BaseFanPWMRead",
    "BaseFanPWMWrite",
    "BaseFanSpeed",
    "FanInputDevice",
    "FanValue",
    "FreeIPMIFanSpeed",
    "LinuxFanPWMRead",
    "LinuxFanPWMWrite",
    "LinuxFanSpeed",
    "PWMDevice",
    "PWMValue",
)



import subprocess
from unittest.mock import patch

import pytest

from afancontrol.temp import HDDTemp, TempCelsius, TempStatus


@pytest.fixture
def hddtemp_output_many():
    return (
        "/dev/sda: Adaptec XXXXX:  drive supported,"
        " but it doesn't have a temperature sensor.\n"
        "/dev/sdb: Adaptec XXXXX:  drive supported,"
        " but it doesn't have a temperature sensor.\n"
        "38\n"
        "39\n"
        "30\n"
        "36\n"
    )


@pytest.fixture
def hddtemp_output_bad():
    return (
        "/dev/sda: Adaptec XXXXX:  drive supported,"
        " but it doesn't have a temperature sensor.\n"
    )


def test_hddtemp_many(hddtemp_output_many):
    with patch.object(HDDTemp, "_call_hddtemp") as mock_call_hddtemp:
        mock_call_hddtemp.return_value = hddtemp_output_many
        t = HDDTemp(
            disk_path="/dev/sd?",
            min=TempCelsius(38.0),
            max=TempCelsius(45.0),
            panic=TempCelsius(50.0),
            threshold=None,
            hddtemp_bin="testbin",
        )

        assert t.get() == TempStatus(
            temp=TempCelsius(39.0),
            min=TempCelsius(38.0),
            max=TempCelsius(45.0),
            panic=TempCelsius(50.0),
            threshold=None,
            is_panic=False,
            is_threshold=False,
        )
        print(repr(t))


def test_hddtemp_bad(hddtemp_output_bad):
    with patch.object(HDDTemp, "_call_hddtemp") as mock_call_hddtemp:
        mock_call_hddtemp.return_value = hddtemp_output_bad
        t = HDDTemp(
            disk_path="/dev/sda",
            min=TempCelsius(38.0),
            max=TempCelsius(45.0),
            panic=TempCelsius(50.0),
            threshold=None,
            hddtemp_bin="testbin",
        )
        with pytest.raises(RuntimeError):
            t.get()


def test_hddtemp_exec_successful(temp_path):
    (temp_path / "sda").write_text("")
    (temp_path / "sdz").write_text("")
    t = HDDTemp(
        disk_path=str(temp_path / "sd") + "?",
        min=TempCelsius(38.0),
        max=TempCelsius(45.0),
        panic=TempCelsius(50.0),
        threshold=None,
        hddtemp_bin="printf '@%s'",
    )
    expected_out = "@-n@-u@C@--@{0}/sda@{0}/sdz".format(temp_path)
    assert expected_out == t._call_hddtemp()


def test_hddtemp_exec_failed():
    t = HDDTemp(
        disk_path="/dev/sd?",
        min=TempCelsius(38.0),
        max=TempCelsius(45.0),
        panic=TempCelsius(50.0),
        threshold=None,
        hddtemp_bin="false",
    )
    with pytest.raises(subprocess.CalledProcessError):
        t._call_hddtemp()



import abc
import contextlib
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn
from timeit import default_timer
from typing import TYPE_CHECKING, Mapping, Optional, Union
from urllib.parse import parse_qs, urlparse

from afancontrol.arduino import ArduinoConnection, ArduinoName
from afancontrol.config import AnyFanName, FanName, ReadonlyFanName, TempName
from afancontrol.fans import Fans
from afancontrol.logger import logger
from afancontrol.pwmfannorm import PWMFanNorm, ReadonlyPWMFanNorm
from afancontrol.temps import ObservedTempStatus
from afancontrol.trigger import Triggers

if TYPE_CHECKING:
    from typing import ContextManager  # Added in 3.6

try:
    import prometheus_client as prom

    prometheus_available = True
except ImportError:
    prometheus_available = False


class Metrics(abc.ABC):
    @abc.abstractmethod
    def __enter__(self):
        pass

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, exc_tb):
        pass

    @abc.abstractmethod
    def tick(
        self,
        temps: Mapping[TempName, ObservedTempStatus],
        fans: Fans,
        triggers: Triggers,
        arduino_connections: Mapping[ArduinoName, ArduinoConnection],
    ) -> None:
        pass

    @abc.abstractmethod
    def measure_tick(self) -> "ContextManager[None]":
        pass


class NullMetrics(Metrics):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass

    def tick(
        self,
        temps: Mapping[TempName, ObservedTempStatus],
        fans: Fans,
        triggers: Triggers,
        arduino_connections: Mapping[ArduinoName, ArduinoConnection],
    ) -> None:
        pass

    def measure_tick(self) -> "ContextManager[None]":
        @contextlib.contextmanager
        def null_context_manager():
            yield

        return null_context_manager()


class PrometheusMetrics(Metrics):
    def __init__(self, listen_host: str) -> None:
        if not prometheus_available:
            raise RuntimeError(
                "`prometheus_client` is not installed. "
                "Run `pip install 'afancontrol[metrics]'`."
            )

        self._listen_addr, port_str = listen_host.rsplit(":", 1)
        self._listen_port = int(port_str)

        self._http_server = None  # type: Optional[HTTPServer]

        self._last_metrics_collect_clock = float("nan")

        # Create a separate registry for this instance instead of using
        # the default one (which is global and doesn't allow to instantiate
        # this class more than once due to having metrics below being
        # registered for a second time):
        self.registry = prom.CollectorRegistry(auto_describe=True)

        # Register some default prometheus_client metrics:
        prom.ProcessCollector(registry=self.registry)
        if hasattr(prom, "PlatformCollector"):
            prom.PlatformCollector(registry=self.registry)
        if hasattr(prom, "GCCollector"):
            prom.GCCollector(registry=self.registry)

        # Temps:
        self.temperature_is_failing = prom.Gauge(
            "temperature_is_failing",
            "The temperature sensor is failing (it isn't returning any data)",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_current = prom.Gauge(
            "temperature_current",
            "The current (filtered) temperature value (in Celsius) "
            "from a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_min = prom.Gauge(
            "temperature_min",
            "The min temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_max = prom.Gauge(
            "temperature_max",
            "The max temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_panic = prom.Gauge(
            "temperature_panic",
            "The panic temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_threshold = prom.Gauge(
            "temperature_threshold",
            "The threshold temperature value (in Celsius) for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_is_panic = prom.Gauge(
            "temperature_is_panic",
            "Is panic temperature reached for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )
        self.temperature_is_threshold = prom.Gauge(
            "temperature_is_threshold",
            "Is threshold temperature reached for a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )

        self.temperature_current_raw = prom.Gauge(
            "temperature_current_raw",
            "The current (unfiltered) temperature value (in Celsius) "
            "from a temperature sensor",
            ["temp_name"],
            registry=self.registry,
        )

        # Fans:
        self.fan_rpm = prom.Gauge(
            "fan_rpm",
            "Fan speed (in RPM) as reported by the fan",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm = prom.Gauge(
            "fan_pwm",
            "Current fan's PWM value (from 0 to 255)",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm_normalized = prom.Gauge(
            "fan_pwm_normalized",
            "Current fan's normalized PWM value (from 0.0 to 1.0, within "
            "the `fan_pwm_line_start` and `fan_pwm_line_end` interval)",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm_line_start = prom.Gauge(
            "fan_pwm_line_start",
            "PWM value where a linear correlation with RPM starts for the fan",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_pwm_line_end = prom.Gauge(
            "fan_pwm_line_end",
            "PWM value where a linear correlation with RPM ends for the fan",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_is_stopped = prom.Gauge(
            "fan_is_stopped",
            "Is PWM fan stopped because the corresponding temperatures "
            "are already low",
            ["fan_name"],
            registry=self.registry,
        )
        self.fan_is_failing = prom.Gauge(
            "fan_is_failing",
            "Is PWM fan marked as failing (e.g. because it has jammed)",
            ["fan_name"],
            registry=self.registry,
        )

        # Arduino boards:
        self.arduino_is_connected = prom.Gauge(
            "arduino_is_connected",
            "Is Arduino board connected via Serial",
            ["arduino_name"],
            registry=self.registry,
        )
        self.arduino_status_age_seconds = prom.Gauge(
            "arduino_status_age_seconds",
            "Seconds since the last `status` message from "
            "the Arduino board (measured at the latest tick)",
            ["arduino_name"],
            registry=self.registry,
        )

        # Others:
        self.is_panic = prom.Gauge(
            "is_panic", "Is in panic mode", registry=self.registry
        )
        self.is_threshold = prom.Gauge(
            "is_threshold", "Is in threshold mode", registry=self.registry
        )

        self.tick_duration = prom.Histogram(
            # Summary would have been better there, but prometheus_client
            # doesn't yet support quantiles in Summaries.
            # See: https://github.com/prometheus/client_python/issues/92
            "tick_duration",
            "Duration of a single tick",
            buckets=(0.1, 0.25, 0.5, 0.75, 1.0, 2.5, 5.0, 10.0, float("inf")),
            registry=self.registry,
        )
        last_metrics_tick_seconds_ago = prom.Gauge(
            "last_metrics_tick_seconds_ago",
            "The time in seconds since the last tick (which also updates these metrics)",
            registry=self.registry,
        )
        last_metrics_tick_seconds_ago.set_function(
            lambda: self.last_metrics_tick_seconds_ago
        )

    @property
    def last_metrics_tick_seconds_ago(self):
        return self._clock() - self._last_metrics_collect_clock

    def _start(self):
        # `prometheus_client.start_http_server` which persists a server reference
        # so it could be stopped later.
        CustomMetricsHandler = MetricsHandler.factory(self.registry)
        httpd = _ThreadingSimpleServer(
            (self._listen_addr, self._listen_port), CustomMetricsHandler
        )
        t = threading.Thread(target=httpd.serve_forever)
        t.daemon = True
        t.start()
        return httpd

    def __enter__(self):
        self._http_server = self._start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self._http_server is not None
        self._http_server.shutdown()  # stop serve_forever()
        self._http_server.server_close()
        self._http_server = None
        return None

    def tick(
        self,
        temps: Mapping[TempName, ObservedTempStatus],
        fans: Fans,
        triggers: Triggers,
        arduino_connections: Mapping[ArduinoName, ArduinoConnection],
    ) -> None:
        for temp_name, observed_temp_status in temps.items():
            temp_status = observed_temp_status.filtered
            if temp_status is None:
                self.temperature_is_failing.labels(temp_name).set(1)
                self.temperature_current.labels(temp_name).set(none_to_nan(None))
                self.temperature_min.labels(temp_name).set(none_to_nan(None))
                self.temperature_max.labels(temp_name).set(none_to_nan(None))
                self.temperature_panic.labels(temp_name).set(none_to_nan(None))
                self.temperature_threshold.labels(temp_name).set(none_to_nan(None))
                self.temperature_is_panic.labels(temp_name).set(none_to_nan(None))
                self.temperature_is_threshold.labels(temp_name).set(none_to_nan(None))
            else:
                self.temperature_is_failing.labels(temp_name).set(0)
                self.temperature_current.labels(temp_name).set(temp_status.temp)
                self.temperature_min.labels(temp_name).set(temp_status.min)
                self.temperature_max.labels(temp_name).set(temp_status.max)
                self.temperature_panic.labels(temp_name).set(
                    none_to_nan(temp_status.panic)
                )
                self.temperature_threshold.labels(temp_name).set(
                    none_to_nan(temp_status.threshold)
                )
                self.temperature_is_panic.labels(temp_name).set(temp_status.is_panic)
                self.temperature_is_threshold.labels(temp_name).set(
                    temp_status.is_threshold
                )

            temp_status = observed_temp_status.raw
            if temp_status is None:
                self.temperature_current_raw.labels(temp_name).set(none_to_nan(None))
            else:
                self.temperature_current_raw.labels(temp_name).set(temp_status.temp)

        for fan_name, pwmfan_norm in fans.fans.items():
            self._collect_fan_metrics(fans, fan_name, pwmfan_norm)
        for readonly_fan_name, readonly_pwmfan_norm in fans.readonly_fans.items():
            self._collect_readonly_fan_metrics(
                fans, readonly_fan_name, readonly_pwmfan_norm
            )

        for arduino_name, arduino_connection in arduino_connections.items():
            self.arduino_is_connected.labels(arduino_name).set(
                arduino_connection.is_connected
            )
            self.arduino_status_age_seconds.labels(arduino_name).set(
                arduino_connection.status_age_seconds
            )

        self.is_panic.set(triggers.panic_trigger.is_alerting)
        self.is_threshold.set(triggers.threshold_trigger.is_alerting)

        self._last_metrics_collect_clock = self._clock()

    def measure_tick(self) -> "ContextManager[None]":
        return self.tick_duration.time()

    def _collect_fan_metrics(
        self, fans: Fans, fan_name: FanName, pwm_fan_norm: PWMFanNorm
    ):
        self.fan_pwm_line_start.labels(fan_name).set(pwm_fan_norm.pwm_line_start)
        self.fan_pwm_line_end.labels(fan_name).set(pwm_fan_norm.pwm_line_end)
        self._collect_any_fan_metrics(fans, fan_name, pwm_fan_norm)

    def _collect_readonly_fan_metrics(
        self, fans: Fans, fan_name: ReadonlyFanName, pwm_fan_norm: ReadonlyPWMFanNorm
    ):
        self._collect_any_fan_metrics(fans, fan_name, pwm_fan_norm)

    def _collect_any_fan_metrics(
        self,
        fans: Fans,
        fan_name: AnyFanName,
        pwm_fan_norm: Union[PWMFanNorm, ReadonlyPWMFanNorm],
    ):
        self.fan_is_stopped.labels(fan_name).set(fans.is_fan_stopped(fan_name))
        self.fan_is_failing.labels(fan_name).set(fans.is_fan_failing(fan_name))
        try:
            self.fan_rpm.labels(fan_name).set(pwm_fan_norm.get_speed())
            self.fan_pwm.labels(fan_name).set(none_to_nan(pwm_fan_norm.get_raw()))
            self.fan_pwm_normalized.labels(fan_name).set(
                none_to_nan(pwm_fan_norm.get())
            )
        except Exception:
            logger.warning(
                "Failed to collect metrics for fan %s", fan_name, exc_info=True
            )
            self.fan_rpm.labels(fan_name).set(none_to_nan(None))
            self.fan_pwm.labels(fan_name).set(none_to_nan(None))
            self.fan_pwm_normalized.labels(fan_name).set(none_to_nan(None))

    def _clock(self):
        return default_timer()


def none_to_nan(v: Optional[float]) -> float:
    if v is None:
        return float("nan")
    return v


class _ThreadingSimpleServer(ThreadingMixIn, HTTPServer):
    """Thread per request HTTP server."""

    # https://github.com/prometheus/client_python/blob/31f5557e2e84ca4ffa9a03abf6e3f4d0c8b8c3eb/prometheus_client/exposition.py#L180-L187  # noqa
    #
    # Make worker threads "fire and forget". Beginning with Python 3.7 this
    # prevents a memory leak because ``ThreadingMixIn`` starts to gather all
    # non-daemon threads in a list in order to join on them at server close.
    # Enabling daemon threads virtually makes ``_ThreadingSimpleServer`` the
    # same as Python 3.7's ``ThreadingHTTPServer``.
    daemon_threads = True


# `MetricsHandler` of `prometheus_client==0.0.18` doesn't support exposing
# a custom registry. This is backported below:
if prometheus_available:
    if hasattr(prom.MetricsHandler, "factory"):
        MetricsHandler = prom.MetricsHandler
    else:
        from prometheus_client.exposition import generate_latest, CONTENT_TYPE_LATEST

        class MetricsHandler(BaseHTTPRequestHandler):  # type: ignore
            # https://github.com/prometheus/client_python/blob/31f5557e2e84ca4ffa9a03abf6e3f4d0c8b8c3eb/prometheus_client/exposition.py#L141-L177  # noqa
            registry = prom.REGISTRY

            def do_GET(self):
                registry = self.registry
                params = parse_qs(urlparse(self.path).query)
                if "name[]" in params:
                    registry = registry.restricted_registry(params["name[]"])
                try:
                    output = generate_latest(registry)
                except Exception:
                    self.send_error(500, "error generating metric output")
                    raise
                self.send_response(200)
                self.send_header("Content-Type", CONTENT_TYPE_LATEST)
                self.end_headers()
                self.wfile.write(output)

            def log_message(self, format, *args):
                """Log nothing."""

            @classmethod
            def factory(cls, registry):
                cls_name = str(cls.__name__)
                MyMetricsHandler = type(cls_name, (cls, object), {"registry": registry})
                return MyMetricsHandler



from typing import Optional, Tuple

from afancontrol.exec import exec_shell_command
from afancontrol.temp.base import Temp, TempCelsius


class CommandTemp(Temp):
    def __init__(
        self,
        shell_command: str,
        *,
        min: Optional[TempCelsius],
        max: Optional[TempCelsius],
        panic: Optional[TempCelsius],
        threshold: Optional[TempCelsius]
    ) -> None:
        super().__init__(panic=panic, threshold=threshold)
        self._shell_command = shell_command
        self._min = min
        self._max = max

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._shell_command == other._shell_command
                and self._min == other._min
                and self._max == other._max
                and self._panic == other._panic
                and self._threshold == other._threshold
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, min=%r, max=%r, panic=%r, threshold=%r)" % (
            type(self).__name__,
            self._shell_command,
            self._min,
            self._max,
            self._panic,
            self._threshold,
        )

    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        temps = [
            float(line.strip())
            for line in exec_shell_command(self._shell_command).split("\n")
            if line.strip()
        ]
        temp = TempCelsius(temps[0])

        if self._min is not None:
            min_t = self._min
        else:
            min_t = TempCelsius(temps[1])

        if self._max is not None:
            max_t = self._max
        else:
            max_t = TempCelsius(temps[2])

        return temp, min_t, max_t



from afancontrol.temp.base import Temp, TempCelsius, TempStatus
from afancontrol.temp.command import CommandTemp
from afancontrol.temp.file import FileTemp
from afancontrol.temp.hdd import HDDTemp

__all__ = (
    "CommandTemp",
    "FileTemp",
    "HDDTemp",
    "Temp",
    "TempCelsius",
    "TempStatus",
)



from unittest.mock import patch

import pytest

from afancontrol.pwmfan import FanValue, FreeIPMIFanSpeed


@pytest.fixture
def ipmi_sensors_output():
    return """
ID,Name,Type,Reading,Units,Event
17,FAN1,Fan,1400.00,RPM,'OK'
18,FAN2,Fan,1800.00,RPM,'OK'
19,FAN3,Fan,N/A,RPM,N/A
20,FAN4,Fan,N/A,RPM,N/A
21,FAN5,Fan,N/A,RPM,N/A
22,FAN6,Fan,N/A,RPM,N/A
""".lstrip()


def test_fan_speed(ipmi_sensors_output):
    fan_speed = FreeIPMIFanSpeed("FAN2")
    with patch.object(FreeIPMIFanSpeed, "_call_ipmi_sensors") as mock_call:
        mock_call.return_value = ipmi_sensors_output
        assert fan_speed.get_speed() == FanValue(1800)


def test_fan_speed_na(ipmi_sensors_output):
    fan_speed = FreeIPMIFanSpeed("FAN3")
    with patch.object(FreeIPMIFanSpeed, "_call_ipmi_sensors") as mock_call:
        mock_call.return_value = ipmi_sensors_output
        with pytest.raises(ValueError):
            fan_speed.get_speed()


def test_fan_speed_unknown(ipmi_sensors_output):
    fan_speed = FreeIPMIFanSpeed("FAN30")
    with patch.object(FreeIPMIFanSpeed, "_call_ipmi_sensors") as mock_call:
        mock_call.return_value = ipmi_sensors_output
        with pytest.raises(RuntimeError):
            fan_speed.get_speed()



import abc
from typing import NamedTuple, NewType, Optional, Tuple

TempCelsius = NewType("TempCelsius", float)

TempStatus = NamedTuple(
    "TempStatus",
    [
        ("temp", TempCelsius),
        ("min", TempCelsius),
        ("max", TempCelsius),
        ("panic", Optional[TempCelsius]),
        ("threshold", Optional[TempCelsius]),
        ("is_panic", bool),
        ("is_threshold", bool),
    ],
)


class Temp(abc.ABC):
    def __init__(
        self, *, panic: Optional[TempCelsius], threshold: Optional[TempCelsius]
    ) -> None:
        self._panic = panic
        self._threshold = threshold

    def get(self) -> TempStatus:
        temp, min_t, max_t = self._get_temp()

        if not (min_t < max_t):
            raise RuntimeError(
                "Min temperature must be less than max. %s < %s" % (min_t, max_t)
            )

        return TempStatus(
            temp=temp,
            min=min_t,
            max=max_t,
            panic=self._panic,
            threshold=self._threshold,
            is_panic=self._panic is not None and temp >= self._panic,
            is_threshold=self._threshold is not None and temp >= self._threshold,
        )

    @abc.abstractmethod
    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        pass



from collections import defaultdict
from contextlib import ExitStack
from typing import Dict, Mapping, Optional

from afancontrol.arduino import ArduinoConnection, ArduinoName
from afancontrol.config import (
    FanName,
    FansTempsRelation,
    MappingName,
    ReadonlyFanName,
    TempName,
    TriggerConfig,
)
from afancontrol.fans import Fans
from afancontrol.logger import logger
from afancontrol.metrics import Metrics
from afancontrol.pwmfannorm import PWMFanNorm, PWMValueNorm, ReadonlyPWMFanNorm
from afancontrol.report import Report
from afancontrol.temp import TempStatus
from afancontrol.temps import FilteredTemp, Temps, filtered_temps
from afancontrol.trigger import Triggers


class Manager:
    def __init__(
        self,
        *,
        arduino_connections: Mapping[ArduinoName, ArduinoConnection],
        fans: Mapping[FanName, PWMFanNorm],
        readonly_fans: Mapping[ReadonlyFanName, ReadonlyPWMFanNorm],
        temps: Mapping[TempName, FilteredTemp],
        mappings: Mapping[MappingName, FansTempsRelation],
        report: Report,
        triggers_config: TriggerConfig,
        metrics: Metrics
    ) -> None:
        self.report = report
        self.arduino_connections = arduino_connections
        self.fans = Fans(fans=fans, readonly_fans=readonly_fans, report=report)
        self.temps = Temps(temps)
        self.mappings = mappings
        self.triggers = Triggers(triggers_config, report)
        self.metrics = metrics
        self._stack = None  # type: Optional[ExitStack]

    def __enter__(self):  # reusable
        self._stack = ExitStack()
        try:
            self._stack.enter_context(self.fans)
            self._stack.enter_context(self.temps)
            self._stack.enter_context(self.triggers)
            self._stack.enter_context(self.metrics)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self._stack is not None
        self._stack.close()
        return None

    def tick(self) -> None:
        with self.metrics.measure_tick():
            temps = self.temps.get_temps()
            _filtered_temps = filtered_temps(temps)
            self.fans.check_speeds()

            self.triggers.check(_filtered_temps)

            if self.triggers.is_alerting:
                self.fans.set_all_to_full_speed()
            else:
                speeds = self._map_temps_to_fan_speeds(_filtered_temps)
                self.fans.set_fan_speeds(speeds)

        try:
            self.metrics.tick(temps, self.fans, self.triggers, self.arduino_connections)
        except Exception:
            logger.warning("Failed to collect metrics", exc_info=True)

    def _map_temps_to_fan_speeds(
        self, temps: Mapping[TempName, Optional[TempStatus]]
    ) -> Mapping[FanName, PWMValueNorm]:

        temp_speeds = {
            temp_name: self._temp_speed(temp_status)
            for temp_name, temp_status in temps.items()
        }

        fan_speeds = defaultdict(
            lambda: PWMValueNorm(0.0)
        )  # type: Dict[FanName, PWMValueNorm]

        for mapping_name, relation in self.mappings.items():
            mapping_speed = max(temp_speeds[temp_name] for temp_name in relation.temps)
            for fan_modifier in relation.fans:
                pwm_norm = PWMValueNorm(mapping_speed * fan_modifier.modifier)
                pwm_norm = max(pwm_norm, PWMValueNorm(0.0))
                pwm_norm = min(pwm_norm, PWMValueNorm(1.0))
                fan_speeds[fan_modifier.fan] = max(
                    pwm_norm, fan_speeds[fan_modifier.fan]
                )

        # Ensure that all fans have been referenced through the mappings.
        # This is also enforced in the `config.py` module.
        assert len(fan_speeds) == len(self.fans.fans)

        return fan_speeds

    def _temp_speed(self, temp: Optional[TempStatus]) -> PWMValueNorm:
        if temp is None:
            # Failing sensor -- this is the panic mode.
            return PWMValueNorm(1.0)
        speed = PWMValueNorm((temp.temp - temp.min) / (temp.max - temp.min))
        speed = max(speed, PWMValueNorm(0.0))
        speed = min(speed, PWMValueNorm(1.0))
        return speed



from pathlib import Path
from typing import NewType

from afancontrol.pwmfan.base import (
    BaseFanPWMRead,
    BaseFanPWMWrite,
    BaseFanSpeed,
    FanValue,
    PWMValue,
)

PWMDevice = NewType("PWMDevice", str)
FanInputDevice = NewType("FanInputDevice", str)


class LinuxFanSpeed(BaseFanSpeed):
    __slots__ = ("_fan_input",)

    def __init__(self, fan_input: FanInputDevice) -> None:
        self._fan_input = Path(fan_input)

    def get_speed(self) -> FanValue:
        return FanValue(int(self._fan_input.read_text()))


class LinuxFanPWMRead(BaseFanPWMRead):
    __slots__ = ("_pwm",)

    max_pwm = PWMValue(255)
    min_pwm = PWMValue(0)

    def __init__(self, pwm: PWMDevice) -> None:
        self._pwm = Path(pwm)

    def get(self) -> PWMValue:
        return PWMValue(int(self._pwm.read_text()))


class LinuxFanPWMWrite(BaseFanPWMWrite):
    __slots__ = "_pwm", "_pwm_enable"

    read_cls = LinuxFanPWMRead

    def __init__(self, pwm: PWMDevice) -> None:
        self._pwm = Path(pwm)
        self._pwm_enable = Path(pwm + "_enable")

    def _set_raw(self, pwm: PWMValue) -> None:
        self._pwm.write_text(str(int(pwm)))

    def __enter__(self):  # reusable
        # fancontrol way of doing it
        if self._pwm_enable.is_file():
            self._pwm_enable.write_text("1")
        self.set_full_speed()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        # fancontrol way of doing it
        if not self._pwm_enable.is_file():
            self.set_full_speed()
            return

        self._pwm_enable.write_text("0")
        if self._pwm_enable.read_text().strip() == "0":
            return

        self._pwm_enable.write_text("1")
        self.set_full_speed()

        if (
            self._pwm_enable.read_text().strip() == "1"
            and int(self._pwm.read_text()) >= self.read_cls.max_pwm
        ):
            return

        raise RuntimeError("Couldn't disable PWM on the fan %r" % self)



import concurrent.futures
from contextlib import ExitStack
from typing import Mapping, NamedTuple, Optional

from afancontrol.config import FilteredTemp, TempName
from afancontrol.filters import TempFilter
from afancontrol.logger import logger
from afancontrol.temp import Temp, TempStatus

ObservedTempStatus = NamedTuple(
    "ObservedTempStatus",
    [("raw", Optional[TempStatus]), ("filtered", Optional[TempStatus])],
)


def filtered_temps(
    temps: Mapping[TempName, ObservedTempStatus]
) -> Mapping[TempName, Optional[TempStatus]]:
    return {
        temp_name: observed_temp_status.filtered
        for temp_name, observed_temp_status in temps.items()
    }


class Temps:
    def __init__(self, temps: Mapping[TempName, FilteredTemp]) -> None:
        self.temps = temps
        self._stack = None  # type: Optional[ExitStack]
        self._executor = None  # type: Optional[concurrent.futures.Executor]

    def __enter__(self):  # reusable
        self._stack = ExitStack()
        try:
            for filtered_temp in self.temps.values():
                self._stack.enter_context(filtered_temp.filter)
            self._executor = self._stack.enter_context(
                concurrent.futures.ThreadPoolExecutor()
            )
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self._stack is not None
        self._stack.close()
        self._executor = None

    def get_temps(self) -> Mapping[TempName, ObservedTempStatus]:
        assert self._executor is not None
        futures = {
            temp_name: self._executor.submit(
                _get_temp_status,
                temp_name,
                temp=filtered_temp.temp,
                filter=filtered_temp.filter,
            )
            for temp_name, filtered_temp in self.temps.items()
        }
        return {temp_name: future.result() for temp_name, future in futures.items()}


def _get_temp_status(
    name: TempName, temp: Temp, filter: TempFilter
) -> ObservedTempStatus:
    try:
        sensor_value = temp.get()  # type: Optional[TempStatus]
    except Exception as e:
        sensor_value = None
        logger.warning("Temp sensor [%s] has failed: %s", name, e, exc_info=True)

    filtered_value = filter.apply(sensor_value)
    logger.debug(
        "Temp status [%s]: actual=%s, filtered=%s", name, sensor_value, filtered_value
    )

    return ObservedTempStatus(raw=sensor_value, filtered=filtered_value)



#!/usr/bin/env python3

import re

from setuptools import setup

with open("src/afancontrol/__init__.py", "rt") as f:
    version = re.search(r'^__version__ = "(.*?)"$', f.read()).group(1)

setup(
    version=version,
    data_files=[
        ("etc/afancontrol", ["pkg/afancontrol.conf"]),
        ("etc/systemd/system", ["pkg/afancontrol.service"]),
    ],
)



import abc
from typing import NewType, Type

PWMValue = NewType("PWMValue", int)  # [0..255]
FanValue = NewType("FanValue", int)


class _SlotsReprMixin:
    def __eq__(self, other):
        if isinstance(other, type(self)):
            for attr in self.__slots__:
                if getattr(self, attr) != getattr(other, attr):
                    return False
            return True

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        # repr assumes that the `__slots__` attrs match the `__init__` signature.

        return "%s(%s)" % (
            type(self).__name__,
            ", ".join(repr(getattr(self, attr)) for attr in self.__slots__),
        )


class BaseFanSpeed(abc.ABC, _SlotsReprMixin):
    @abc.abstractmethod
    def get_speed(self) -> FanValue:
        pass

    def __enter__(self):  # reusable
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass


class BaseFanPWMRead(abc.ABC, _SlotsReprMixin):
    max_pwm = None  # type: PWMValue
    min_pwm = None  # type: PWMValue

    def is_stopped(self) -> bool:
        return type(self).is_pwm_stopped(self.get())

    @staticmethod
    def is_pwm_stopped(pwm: PWMValue) -> bool:
        return pwm <= 0

    @abc.abstractmethod
    def get(self) -> PWMValue:
        pass

    def __enter__(self):  # reusable
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass


class BaseFanPWMWrite(abc.ABC, _SlotsReprMixin):
    read_cls = None  # type: Type[BaseFanPWMRead]

    def set(self, pwm: PWMValue) -> None:
        if not (self.read_cls.min_pwm <= pwm <= self.read_cls.max_pwm):
            raise ValueError(
                "Invalid pwm value %s: it must be within [%s..%s]"
                % (pwm, self.read_cls.min_pwm, self.read_cls.max_pwm)
            )
        self._set_raw(pwm)

    def set_full_speed(self) -> None:
        self._set_raw(self.read_cls.max_pwm)

    @abc.abstractmethod
    def _set_raw(self, pwm: PWMValue) -> None:
        pass

    def __enter__(self):  # reusable
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        pass



from afancontrol.arduino import ArduinoConnection, ArduinoPin
from afancontrol.pwmfan.base import (
    BaseFanPWMRead,
    BaseFanPWMWrite,
    BaseFanSpeed,
    FanValue,
    PWMValue,
)


class ArduinoFanSpeed(BaseFanSpeed):
    __slots__ = "_conn", "_tacho_pin"

    def __init__(
        self, arduino_connection: ArduinoConnection, *, tacho_pin: ArduinoPin
    ) -> None:
        self._conn = arduino_connection
        self._tacho_pin = tacho_pin

    def get_speed(self) -> FanValue:
        return FanValue(self._conn.get_rpm(self._tacho_pin))

    def __enter__(self):  # reusable
        self._conn.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._conn.__exit__(exc_type, exc_value, exc_tb)


class ArduinoFanPWMRead(BaseFanPWMRead):
    __slots__ = "_conn", "_pwm_pin"

    max_pwm = PWMValue(255)
    min_pwm = PWMValue(0)

    def __init__(
        self, arduino_connection: ArduinoConnection, *, pwm_pin: ArduinoPin
    ) -> None:
        self._conn = arduino_connection
        self._pwm_pin = pwm_pin

    def get(self) -> PWMValue:
        return PWMValue(int(self._conn.get_pwm(self._pwm_pin)))

    def __enter__(self):  # reusable
        self._conn.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._conn.__exit__(exc_type, exc_value, exc_tb)


class ArduinoFanPWMWrite(BaseFanPWMWrite):
    __slots__ = "_conn", "_pwm_pin"

    read_cls = ArduinoFanPWMRead

    def __init__(
        self, arduino_connection: ArduinoConnection, *, pwm_pin: ArduinoPin
    ) -> None:
        self._conn = arduino_connection
        self._pwm_pin = pwm_pin

    def _set_raw(self, pwm: PWMValue) -> None:
        self._conn.set_pwm(self._pwm_pin, pwm)

    def __enter__(self):  # reusable
        self._conn.__enter__()
        self.set_full_speed()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        try:
            self.set_full_speed()
            self._conn.wait_for_status()

            if int(self._conn.get_pwm(self._pwm_pin)) >= self.read_cls.max_pwm:
                return

            raise RuntimeError("Couldn't disable PWM on the fan %r" % self)
        finally:
            self._conn.__exit__(exc_type, exc_value, exc_tb)



import glob
import re
from pathlib import Path
from typing import Optional, Tuple

from afancontrol.temp.base import Temp, TempCelsius


def _expand_glob(path: str):
    matches = glob.glob(path)
    if not matches:
        return path  # a FileNotFoundError will be raised on a first read attempt
    if len(matches) == 1:
        return matches[0]
    raise ValueError("Expected glob to expand to a single path, got %r" % (matches,))


class FileTemp(Temp):
    def __init__(
        self,
        temp_path: str,  # /sys/class/hwmon/hwmon0/temp1
        *,
        min: Optional[TempCelsius],
        max: Optional[TempCelsius],
        panic: Optional[TempCelsius],
        threshold: Optional[TempCelsius]
    ) -> None:
        super().__init__(panic=panic, threshold=threshold)
        temp_path = re.sub(r"_input$", "", temp_path)

        # Allow paths looking like this (this one is from an nvme drive):
        #  /sys/devices/pci0000:00/0000:00:01.3/[...]/hwmon/hwmon*/temp1_input
        # The `hwmon*` might change after reboot, but it is always a single
        # directory within the device.
        temp_path = _expand_glob(temp_path + "_input")
        temp_path = re.sub(r"_input$", "", temp_path)

        self._temp_input = Path(temp_path + "_input")
        self._temp_min = Path(temp_path + "_min")
        self._temp_max = Path(temp_path + "_max")
        self._min = min
        self._max = max

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self._temp_input == other._temp_input
                and self._temp_min == other._temp_min
                and self._temp_max == other._temp_max
                and self._min == other._min
                and self._max == other._max
                and self._panic == other._panic
                and self._threshold == other._threshold
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, min=%r, max=%r, panic=%r, threshold=%r)" % (
            type(self).__name__,
            str(self._temp_input),
            self._min,
            self._max,
            self._panic,
            self._threshold,
        )

    def _get_temp(self) -> Tuple[TempCelsius, TempCelsius, TempCelsius]:
        temp = self._read_temp_from_path(self._temp_input)
        return temp, self._get_min(), self._get_max()

    def _get_min(self) -> TempCelsius:
        if self._min is not None:
            return self._min
        try:
            min_t = self._read_temp_from_path(self._temp_min)
        except FileNotFoundError:
            raise RuntimeError(
                "Please specify `min` and `max` temperatures for "
                "the %s sensor" % self._temp_input
            )
        return min_t

    def _get_max(self) -> TempCelsius:
        if self._max is not None:
            return self._max
        try:
            max_t = self._read_temp_from_path(self._temp_max)
        except FileNotFoundError:
            raise RuntimeError(
                "Please specify `min` and `max` temperatures for "
                "the %s sensor" % self._temp_input
            )
        return max_t

    def _read_temp_from_path(self, path: Path) -> TempCelsius:
        return TempCelsius(int(path.read_text().strip()) / 1000)



import math
from contextlib import ExitStack
from typing import NewType, Optional

from afancontrol.pwmfan import (
    BaseFanPWMRead,
    BaseFanPWMWrite,
    BaseFanSpeed,
    FanValue,
    PWMValue,
)

PWMValueNorm = NewType("PWMValueNorm", float)  # [0..1]


class ReadonlyPWMFanNorm:
    def __init__(
        self, fan_speed: BaseFanSpeed, pwm_read: Optional[BaseFanPWMRead] = None
    ) -> None:
        self.fan_speed = fan_speed
        self.pwm_read = pwm_read
        self._stack = None  # type: Optional[ExitStack]

    def __enter__(self):
        self._stack = ExitStack()
        try:
            self._stack.enter_context(self.fan_speed)
            if self.pwm_read is not None:
                self._stack.enter_context(self.pwm_read)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self._stack is not None
        self._stack.close()

    def get_speed(self) -> FanValue:
        return self.fan_speed.get_speed()

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.fan_speed == other.fan_speed and self.pwm_read == other.pwm_read

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, %r)" % (type(self).__name__, self.fan_speed, self.pwm_read)

    def is_pwm_stopped(self, pwm: Optional[PWMValue]) -> Optional[bool]:
        if self.pwm_read is None:
            return None
        if pwm is None:
            return None
        return type(self.pwm_read).is_pwm_stopped(pwm)

    def get_raw(self) -> Optional[PWMValue]:
        if self.pwm_read is None:
            return None
        return self.pwm_read.get()

    def get(self) -> Optional[PWMValueNorm]:
        if self.pwm_read is None:
            return None
        raw = self.get_raw()
        assert raw is not None
        return PWMValueNorm(raw / self.pwm_read.max_pwm)


class PWMFanNorm:
    def __init__(
        self,
        fan_speed: BaseFanSpeed,
        pwm_read: BaseFanPWMRead,
        pwm_write: BaseFanPWMWrite,
        *,
        pwm_line_start: PWMValue,
        pwm_line_end: PWMValue,
        never_stop: bool = False
    ) -> None:
        self.fan_speed = fan_speed
        self.pwm_read = pwm_read
        self.pwm_write = pwm_write
        self.pwm_line_start = pwm_line_start
        self.pwm_line_end = pwm_line_end
        self.never_stop = never_stop
        if type(self.pwm_read).min_pwm > self.pwm_line_start:
            raise ValueError(
                "Invalid pwm_line_start. Expected: min_pwm <= pwm_line_start. "
                "Got: %s <= %s" % (type(self.pwm_read).min_pwm, self.pwm_line_start)
            )
        if self.pwm_line_end > type(self.pwm_read).max_pwm:
            raise ValueError(
                "Invalid pwm_line_end. Expected: pwm_line_end <= max_pwm. "
                "Got: %s <= %s" % (self.pwm_line_end, type(self.pwm_read).max_pwm)
            )
        self._stack = None  # type: Optional[ExitStack]

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.fan_speed == other.fan_speed
                and self.pwm_read == other.pwm_read
                and self.pwm_write == other.pwm_write
                and self.pwm_line_start == other.pwm_line_start
                and self.pwm_line_end == other.pwm_line_end
                and self.never_stop == other.never_stop
            )

        return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "%s(%r, %r, %r, pwm_line_start=%r, pwm_line_end=%r, never_stop=%r)" % (
            type(self).__name__,
            self.fan_speed,
            self.pwm_read,
            self.pwm_write,
            self.pwm_line_start,
            self.pwm_line_end,
            self.never_stop,
        )

    def __enter__(self):
        self._stack = ExitStack()
        try:
            self._stack.enter_context(self.fan_speed)
            self._stack.enter_context(self.pwm_read)
            self._stack.enter_context(self.pwm_write)
        except Exception:
            self._stack.close()
            raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self._stack is not None
        self._stack.close()

    def get_speed(self) -> FanValue:
        return self.fan_speed.get_speed()

    def is_pwm_stopped(self, pwm: PWMValue) -> bool:
        return type(self.pwm_read).is_pwm_stopped(pwm)

    def set_full_speed(self) -> None:
        self.pwm_write.set_full_speed()

    def get_raw(self) -> PWMValue:
        return self.pwm_read.get()

    def get(self) -> PWMValueNorm:
        return PWMValueNorm(self.get_raw() / self.pwm_read.max_pwm)

    def set(self, pwm_norm: PWMValueNorm) -> PWMValue:
        # TODO validate this formula
        pwm_norm = max(pwm_norm, PWMValueNorm(0.0))
        pwm_norm = min(pwm_norm, PWMValueNorm(1.0))
        pwm = pwm_norm * self.pwm_line_end
        if 0 < pwm < self.pwm_line_start:
            pwm = self.pwm_line_start
        if pwm <= 0 and self.never_stop:
            pwm = self.pwm_line_start
        if pwm_norm >= 1.0:
            pwm = self.pwm_read.max_pwm

        pwm = PWMValue(int(math.ceil(pwm)))
        self.pwm_write.set(pwm)
        return pwm
